#!/usr/bin/env python3
###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

__author__      = "Joel Boyd"
__copyright__   = "Copyright 2017"
__credits__     = ["Joel Boyd"]
__license__     = "GPL3"
__version__     = "0.0.7"
__maintainer__  = "Joel Boyd"
__email__       = "joel.boyd near uq.net.au"
__status__      = "Development"

###############################################################################
# Imports
import logging
import os
import random
import re
import numpy as np
import multiprocessing as mp
import statsmodels.sandbox.stats.multicomp as sm
from scipy import stats
from itertools import product, combinations, chain
from collections import Counter
# Local
from enrichm.parse_annotate import ParseAnnotate
from enrichm.module_description_parser import ModuleDescription
from enrichm.databases import Databases
from enrichm.draw_plots import Plot
from enrichm.comparer import Compare
################################################################################

def gene_fisher_calc(x):
    
    annotation, group_1, group_2 = x[0], x[1], x[2]
    
    dat = x[3:]
    
    if (dat[0][0]>0 or dat[1][0]>0) and (dat[0][1]>0 or dat[1][1]>0):
        score, pval = stats.fisher_exact(dat)
    else:
        score, pval = 'nan', 1.0

    return [annotation, group_1, group_2] + dat[0] + dat[1] + [score, pval]

def mannwhitneyu_calc(x):
    # Mann Whitney U test
    annotation, group_1, group_2, group_1_module_annotations, group_2_module_annotations = x
    group_1_module_annotations = group_1_module_annotations[0]
    group_2_module_annotations = group_2_module_annotations[0]

    if(sum(group_1_module_annotations)>0 and sum(group_2_module_annotations)>0):
        if(len(set(group_1_module_annotations)) == 1
            or
           len(set(group_2_module_annotations)) == 1):
            mw_t_stat, mw_p_value = 'NA', 1
        else:
            group_1_module_annotations = np.array(group_1_module_annotations)       
            group_2_module_annotations = np.array(group_2_module_annotations)           
            mw_t_stat, mw_p_value = \
                stats.mannwhitneyu(group_1_module_annotations,
                                         group_2_module_annotations)            
    else:
        mw_t_stat, mw_p_value = 'NA', 1
    return [annotation, group_1, group_2, str(np.mean(group_1_module_annotations)),
            str(np.mean(group_2_module_annotations)), mw_t_stat, mw_p_value]


def zscore_calc(x):
    
    annotation, group_1, group_2, group_1_module_annotations, group_2_module_annotations = x
    group_1_module_annotations = group_1_module_annotations[0]
    group_2_module_annotations = group_2_module_annotations[0]

    if len(group_1_module_annotations)>1:
        reference = group_1_module_annotations
        reference_name = group_1
        genome = group_2_module_annotations[0]
        genome_name = group_2
    else:
        reference_name = group_2
        reference = group_2_module_annotations
        genome = group_1_module_annotations[0]
        genome_name = group_1

    if genome>0:
        reference_group_comp_sd \
            = np.std(reference, axis=0)
        reference_group_comp_mean \
            = np.mean(reference, axis=0)

        if (genome-reference_group_comp_mean)>0:
            if reference_group_comp_sd==0:
                z_score = np.inf
                p_value = 0.0
            else:
                z_score = (genome-reference_group_comp_mean) / reference_group_comp_sd
                p_value = 2-2*stats.norm.cdf(z_score)

            return [annotation,
                    reference_name,
                    genome_name,
                    str(reference_group_comp_mean),
                    str(reference_group_comp_sd),
                    str(genome),
                    str(z_score), 
                    p_value]

################################################################################

class Enrichment:
    
    TIGRFAM                 = "tigrfam"
    PFAM                    = "pfam"
    KEGG                    = "kegg"
    CAZY                    = "cazy"
    EC                      = "ec"
    def __init__(self):

        self.TIGRFAM_PREFIX          = 'TIGR'
        self.PFAM_PREFIX             = 'PF'
        self.KEGG_PREFIX             = 'K'
        self.CAZY_PREFIX             = ["GH", "AA", "GT","PL","CE","CBM", "SLH", "dockerin", "cohesin", "GTCellulosesynt"]
        self.EC_PREFIX             = ["1", "2", "3","4","5","6", "7"]
        self.PROPORTIONS             = 'proportions.tsv'
        self.MODULE_COMPLETENESS     = 'modules.tsv'
        self.UNIQUE_TO_GROUPS        = 'unique_to_groups.tsv'
        self.taxonomy_index_dictionary = {"d__":0, "p__":1, "c__":2, "o__":3, "f__":4, "g__":5, "s__":6}    

    def _parse_matrix(self, matrix_file_io, colnames):
        for line in matrix_file_io:
            sline = line.strip().split('\t')
            rowname, entries = sline[0], sline[1:]
            for colname, entry in zip(colnames, entries):
                yield colname, entry, rowname

    def parse_metadata_matrix(self, matrix_path):
        '''        
        Parameters
        ----------
        matrix_path : String. Path to file containing a matrix of genome rownames.        
        '''

        matrix_file_io  = open(matrix_path)
        cols_to_rows    = dict()
        nr_values       = set()
        attribute_dict  = dict()
        
        for line in matrix_file_io:
            rowname, entry = line.strip().split('\t')   
            nr_values.add(entry)

            if entry in attribute_dict:
                attribute_dict[entry].add(rowname)
            else:
                attribute_dict[entry] = set([rowname])

            if rowname not in cols_to_rows:
                cols_to_rows[rowname] = set([entry])
            else:
                cols_to_rows[rowname].add(entry)

        return cols_to_rows, nr_values, attribute_dict

    def _parse_annotation_matrix(self, annotation_matrix):
        '''        
        Parameters
        ----------
        matrix_path : String. Path to file containing a matrix of genome rownames.        
        '''

        matrix_file_io  = open(annotation_matrix)
        colnames        = matrix_file_io.readline().strip().split('\t')[1:]
        cols_to_rows    = {genome_name:dict() for genome_name in colnames}
        rownames        = set()

        for genome_name, entry, rowname \
                        in self._parse_matrix(matrix_file_io, colnames):
            
            rownames.add(rowname)
            if float(entry) > 0:
                cols_to_rows[genome_name][rowname] = int(entry)

        return cols_to_rows, rownames, colnames

    def check_annotation_type(self, annotations):
        '''
        Takes a random sample of the rownames from the input matrix
        and based on the characters they start with, determines the 
        type of annotation being used.
        
        Parameters
        ----------
        annotaitons     - List. A list of strings, each a rowname 
                          from the original input matrix

        Output
        ------
        The annotation type
        '''

        sample = random.sample(annotations, 1)[0]
        
        cazy_prefix = ''
        for character in list(sample):
            if character.isdigit()!=True:
                if character!='_':
                    cazy_prefix+=character

        if sample.startswith(self.TIGRFAM_PREFIX):
            return self.TIGRFAM
        elif sample.startswith(self.KEGG_PREFIX):
            return self.KEGG
        elif sample.startswith(self.PFAM_PREFIX):
            return self.PFAM
        elif cazy_prefix in self.CAZY_PREFIX:
            return self.CAZY
        elif sample.split('.')[0] in self.EC_PREFIX:
            return self.EC

    def calculate_portions(self, modules, combination_dict, annotations_dict, genome_list, proportions_cutoff):
        '''
        Calculates the portions of genome

        Parameters
        ----------
        modules             - List. List of all possible annotations for the given annotation
                              type (eg, all ko ids, or all pfam ids). 
        combination_dict    - Dictionary. Metadata dictionary, with the groups as keys, and lists
                              of genomes as entries
        annotations_dict    - Dictionary. Annotation dictionary, with the the genome ids as keys,
                              and a list of annotations as the entry for each.  
        genome_list         - List. List of strings, each one a genome name
        proportions_cutoff  - Float. Value with which to cutoff
        '''
        ### ~ TODO: Add a portion of genome column too?

        raw_proportions_output_lines        = [['Module'] + list(combination_dict.keys())]

        for module in modules:
            
            module_values               = dict()
            raw_proportions_output_line = [module]
            
            for group_name, genome_list in combination_dict.items():
                if len(genome_list)>0:
                    coverage = len([genome for genome in genome_list 
                                    if module in annotations_dict[genome]])
                    total    = float(len(genome_list))
                    entry    = coverage/total
                    module_values[group_name] = entry
                    raw_proportions_output_line.append(str(entry))
                else:
                    raw_proportions_output_line.append('0.0')
            
            if max(module_values.values())>0:
                groups = combination_dict.keys()
                for group in groups:
                    group_val = module_values[group]
                    compare_groups = list()
                    if group_val>0:
                        for other_group in groups:
                            if other_group!=group:
                                other_group_val = module_values[other_group]
                                if other_group_val > 0:
                                    compare_value = group_val/other_group_val
                                else:
                                    compare_value = float("inf")
                                compare_groups.append(compare_value)
                    
                    if all([x>proportions_cutoff for x in compare_groups]):                    
                        pass 

            raw_proportions_output_lines.append(raw_proportions_output_line)

        return raw_proportions_output_lines

    def _write(self, output_lines_list, output_path):
        '''
        
        Parameters
        ----------
        output_lines_list   - List. A list of lists. Each sublist is a line, 
                              where each entry is a column entry
        output_path         - String. Path to write output lines to.
        Output
        ------
        '''
        logging.info("Writing results to file: %s" % output_path)

        with open(output_path, 'w') as out_io:
            for line in output_lines_list:
                string = '\t'.join([str(x) for x in line]) + '\n'
                out_io.write(string)
                
    def parse_genomes_to_compare(self, genomes_to_compare_with_group_file):
        genomes_to_compare = set()
        
        for genome in open(genomes_to_compare_with_group_file):
            genome = genome.strip()
            genomes_to_compare.add(genome)
        
        return genomes_to_compare

    def parse_gtdb_matrix(self, genomes, matrix):
        '''
        description
        
        Inputs
        ------
        
        Outputs
        -------
        
        '''
        logging.info('Parsing GTDB matrix')
        
        genomes = list(genomes)
        matrix_io = open(matrix)
        header = matrix_io.readline().strip().split('\t')
        
        indexes = list()
        include = list()
        for genome in genomes:
            if genome in header:
                indexes.append(header.index(genome))
                include.append(genome)
        genomes = include

        output_dict = {genome:dict() for genome in genomes}
        for row in matrix_io:
            srow = row.strip().split()
            annotation = srow[0]
            for genome, index in zip(genomes, indexes):
                count = int(srow[index])
                if count > 0:
                    output_dict[genome][annotation] = int(srow[index])

        return output_dict, genomes

    def do(# Input options
           self, annotate_output, metadata_path, input_modules, abundances, 
           # Runtime options
           genomes_to_compare_with_group_file, pval_cutoff, proportions_cutoff, 
           threshold, multi_test_correction, batchfile, processes,
           ko, pfam, tigrfam, hypothetical, cazy, ec,
           # Output options
           output_directory):

        p  = Plot()
        d  = Databases()
        
        if genomes_to_compare_with_group_file:
            self.genomes_to_compare_with_group = self.parse_genomes_to_compare(genomes_to_compare_with_group_file)
        else:
            self.genomes_to_compare_with_group = None

        logging.info('Parsing annotate output: %s' % (annotate_output))
        pa = ParseAnnotate(annotate_output, processes)
        
        logging.info('Parsing annotations')
        if ko:
            annotation_matrix = pa.ko
            gtdb_annotation_matrix = d.GTDB_KO
        elif pfam:
            annotation_matrix = pa.pfam
            gtdb_annotation_matrix = d.GTDB_PFAM
        elif tigrfam:
            annotation_matrix = pa.tigrfam
            gtdb_annotation_matrix = d.GTDB_TIGRFAM
        elif hypothetical:
            annotation_matrix = pa.hypothetical_cluster
            gtdb_annotation_matrix = None
        elif cazy:
            annotation_matrix = pa.cazy
            gtdb_annotation_matrix = d.GTDB_CAZY
        elif ec:
            annotation_matrix = pa.ec
            gtdb_annotation_matrix = d.GTDB_EC
        
        annotations_dict, modules, genomes \
                    = self._parse_annotation_matrix(annotation_matrix)

        if input_modules:
            logging.info('Limiting to %i modules' % len(modules))
            modules = input_modules

        logging.info('Parsing metadata')
        metadata, metadata_value_lists, attribute_dict \
                    = self.parse_metadata_matrix(metadata_path)

        if batchfile:
            genomes_set = set()
            batchfile_metadata, batchfile_metadata_value_lists, batchfile_attribute_dict \
                        = self.parse_metadata_matrix(batchfile)
            genomes_set = genomes_set.union(set(batchfile_metadata.keys()))
            reference_genome_annotations, genomes_set = self.parse_gtdb_matrix(genomes_set, gtdb_annotation_matrix)

            annotations_dict.update(reference_genome_annotations)
            new_batchfile_attribute_dict = dict()
            for x,y in batchfile_attribute_dict.items():
                s = [z for z in y if z in genomes_set]
                if len(s)>0:
                    new_batchfile_attribute_dict[x] = s
            attribute_dict.update(new_batchfile_attribute_dict)
            batchfile_metadata={x:batchfile_metadata[x] for x in genomes_set}
            metadata.update(batchfile_metadata)
            batchfile_metadata_value_lists = set(new_batchfile_attribute_dict.keys())
            metadata_value_lists = metadata_value_lists.union(batchfile_metadata_value_lists)
        
        logging.info("Comparing sets of genomes")        
        combination_dict = dict()

        for combination in product(*list([metadata_value_lists])):
            genome_list = list()
        
            for genome, attributes in metadata.items():
        
                for feature in combination:
        
                    if feature in attributes:
                        genome_list.append(genome)  
        
            combination_dict['_'.join(combination)] = genome_list

        annotation_type = self.check_annotation_type(modules)

        t = Test(annotations_dict,
                 modules,
                 genomes,
                 combination_dict,
                 annotation_type,
                 threshold,
                 multi_test_correction,
                 pval_cutoff, 
                 processes,
                 d)
        results = t.do(attribute_dict)

        for result in results:
            test_result_lines, test_result_output_file = result

            test_result_output_path = os.path.join(output_directory,
                                                   test_result_output_file)
            self._write(test_result_lines, test_result_output_path)

        raw_portions_path \
            = os.path.join(output_directory, self.PROPORTIONS)
        #unique_to_groups_path \
        #    = os.path.join(output_directory, self.UNIQUE_TO_GROUPS)
        raw_proportions_output_lines \
            = self.calculate_portions(modules, combination_dict, annotations_dict, genome_list, proportions_cutoff)

        self._write(raw_proportions_output_lines, raw_portions_path)

        logging.info('Generating summary plots')

        if annotation_type==self.KEGG:
            logging.info('Finding module completeness in differentially abundant KOs')
            
            for result_file in os.listdir(output_directory):
            
                if(result_file.endswith("fisher.tsv") or result_file.endswith("cdf.tsv")):
                    p.draw_barplots(os.path.join(output_directory, result_file), pval_cutoff, output_directory)
                    
                    g1_sig_kos = set()
                    g2_sig_kos = set()
    
                    result_file_io = open(os.path.join(output_directory, result_file))
                    result_file_io.readline()
                    for line in result_file_io:
                        sline = line.strip().split('\t')
                        if float(sline[-2])<pval_cutoff:
                            if result_file.endswith("fisher.tsv"):
                                g1 = float(sline[3]) / (int(sline[3]) + int(sline[4]))
                                g2 = float(sline[5]) / (int(sline[5]) + int(sline[6]))
                            elif result_file.endswith("cdf.tsv"):
                                g1 = float(sline[3])
                                g2 = float(sline[5])
                            if g1>g2:
                                g1_sig_kos.add(sline[0])
                            else:
                                g2_sig_kos.add(sline[0])
    
                    module_output = [["Module", "Lineage", "Total steps", "Steps covered", "Percentage covered", "Module description"]]
                    for module, definition in d.m2def.items():
                        if module not in d.signature_modules:
                            pathway = ModuleDescription(definition)
                            num_all         = pathway.num_steps()
                            g1_num_covered, _, _, _ = pathway.num_covered_steps(g1_sig_kos)
                            g1_perc_covered    = g1_num_covered / float(num_all)
        
                            g2_num_covered, _, _, _ = pathway.num_covered_steps(g2_sig_kos)
                            g2_perc_covered    = g2_num_covered / float(num_all)
                            if g1_perc_covered>0:
                                output_line = [module, sline[1], num_all, g1_num_covered, g1_perc_covered, d.m[module]]
                                module_output.append(output_line)
                            if g2_perc_covered>0:
                                output_line = [module, sline[2], num_all, g2_num_covered, g2_perc_covered, d.m[module]]
                                module_output.append(output_line)
        
                    prefix = '_vs_'.join([sline[1], sline[2]]).replace(' ', '_')
                    self._write(module_output, os.path.join(output_directory, prefix +'_'+ self.MODULE_COMPLETENESS))   

        p.draw_pca_plot(annotation_matrix, metadata_path, output_directory)

class Test(Enrichment):

    def __init__(self, genome_annotations, modules, genomes, groups, 
                 annotation_type, threshold, multi_test_correction, 
                 pval_cutoff, processes, d):
        '''
        
        Parameters
        ----------
        
        Output
        ------
        '''
        self.FISHER_HEADER = [['annotation',
                               'group_1',
                               'group_2',
                               'group_1_true',
                               'group_1_false',
                               'group_2_true',
                               'group_2_false',
                               'score',
                               'pvalue',
                               'corrected_pvalue',
                               'description']]

        self.MANNWHITNEYU_HEADER =[['annotation',
                                    'group_1',
                                    'group_2',
                                    'group_1_mean',
                                    'group_2_mean',
                                    'score',
                                    'pvalue',
                                    'corrected_pvalue',
                                    'description']]
        
        self.ZSCORE_HEADER = [['annotation',
                               'group_1',
                               'group_2',
                               'group_1_mean',
                               'group_1_sd',
                               'group_2_count',
                               'score',
                               'pvalue',
                               'corrected_pvalue',
                               'description']]
        
        self.PA                     = 'presence_absence'
        self.IVG_OUTPUT             = 'ivg_results.cdf.tsv'
        self.GENE_FISHER_OUTPUT     = 'gvg_results.fisher.tsv'
        self.GVG_OUTPUT             = 'gvg_results.mannwhitneyu.tsv'

        self.threshold              = threshold
        self.multi_test_correction  = multi_test_correction
        self.m2def                  = d.m2def
        self.m                      = d.m
        self.clan2pfam              = d.clan2pfam
        self.clan_to_description    = d.clan2name
        self.k                      = d.k
        self.modules                = modules
        self.tigrfamdescription     = d.tigrfamdescription
        self.pfam2description       = d.pfam2description
        self.ec2description       = d.ec2description
        self.genomes                = genomes
        self.annotation_type        = annotation_type
        self.groups                 = groups
        self.pval_cutoff            = pval_cutoff
        self.pool                   = mp.Pool(processes = processes)

        self.mtc_dict = {'b': 'Bonferroni',
                         's': 'Sidak',
                         'h': 'Holm',
                         'hs': 'Holm-Sidak',
                         'sh': 'Simes-Hochberg',
                         'ho': 'Hommel',
                         'fdr_bh': 'FDR Benjamini-Hochberg',
                         'fdr_by': 'FDR Benjamini-Yekutieli',
                         'fdr_tsbh': 'FDR 2-stage Benjamini-Hochberg',
                         'fdr_tsbky': 'FDR 2-stage Benjamini-Krieger-Yekutieli',
                         'fdr_gbs': 'FDR adaptive Gavrilov-Benjamini-Sarkar'}

        if annotation_type==Enrichment.PFAM:
            self.genome_annotations = dict()
            for key, item in genome_annotations.items():                
                self.genome_annotations[key] = {key.split('.')[0]:entry for key,entry in item.items()}
        else:
            self.genome_annotations = genome_annotations

        self.pfam    = dict()
        self.id2pfam = dict()
        
        for line in open(d.PFAM_CLAN_DB):
            sline       = line.strip().split()
            pfam        = sline[0]
            id          = sline[2]
            description = "%s; %s" % (id, ' '.join(sline[2:]))
            self.pfam[pfam]  = description
            self.id2pfam[id] = pfam
    
    def test_chooser(self, groups):
        groups = [len(x) for x in groups]
        
        # Enrichment
        if any(group == 1 for group in groups):
            enrichment_test = self.PA
            overrepresentation_test = stats.norm.cdf
        else:
            enrichment_test = stats.fisher_exact
            overrepresentation_test = stats.mannwhitneyu
        
        return enrichment_test, overrepresentation_test

    def correct_multi_test(self, pvalues):
        logging.info('Applying multi-test correction using the %s method' % (self.mtc_dict[self.multi_test_correction]) )
        corrected_pvals \
            = sm.multipletests(pvalues,
                               alpha        = self.threshold,
                               method       = self.multi_test_correction,
                               returnsorted = False,
                               is_sorted    = False)[1]        

        return corrected_pvals

    def _strip_kegg_definitions(self, definition):
        kos_list = [ko for ko in re.split("[^\w]", definition) if ko]
        return set(kos_list)

    def gather_genome_annotations(self, group, target_annotations):
        
        genome_annotation_list = list()
        
        for genome in self.groups[group]:
            genome_target_list = self.genome_annotations[genome].intersection(target_annotations)
            genome_annotation_list.append(len(genome_target_list))
        
        return genome_annotation_list

    def get_annotations(self):

        if self.annotation_type==Enrichment.TIGRFAM:
            logging.warning('Comparisons are not possible with TIGRFAM because heirarchical classifications are needed (like in KEGG, COG or PFAM).')
        elif self.annotation_type==Enrichment.KEGG:
            iterator = self.m2def
            annotation_description = self.m
        elif self.annotation_type==Enrichment.PFAM:
            iterator = self.clan2pfam
            annotation_description = self.clan_to_description
        return iterator, annotation_description


    def _count(self, annotation, group, freq):
        if freq:
            group_true = list()

        else:
            group_true = 0
        
        group_false = 0

        for genome in self.groups[group]:
        
            if annotation in self.genome_annotations[genome]:
        
                if freq:
                    group_true.append(self.genome_annotations[genome][annotation])
        
                else:
                    group_true+=1
        
            else:
        
                if freq:
                    group_true.append(0)

                else:
                    group_false+=1

        return group_true, group_false
        
    def gene_frequencies(self, group_1, group_2, freq=False):

        res_list    = list()
        annotations = set(chain(*self.genome_annotations.values()))

        for annotation in annotations:
            passed = True
            group_1_true, group_1_false \
                = self._count(annotation,
                              group_1,
                              freq)
            group_2_true, group_2_false \
                = self._count(annotation,
                              group_2,
                              freq)

            if freq:
                if(len([x for x in group_1_true if x!='0'])==0 and
                    len([x for x in group_2_true if x!='0'])==0 ):
                    passed = False
            else:
                if(group_1_true==0 and group_2_true==0):
                    passed = False
            
            if passed:
                res_list.append([annotation, group_1, group_2, [group_1_true, group_1_false], [group_2_true, group_2_false]])

        return res_list

    def corrected_pvals(self, output_lines):
        pvalues = [output_line[-1] for output_line in output_lines]
        corrected_pvalues = self.correct_multi_test(pvalues)
        return corrected_pvalues

    def add_descriptions(self, output_lines):
        if self.annotation_type == Enrichment.KEGG:
            desc = self.k

        if self.annotation_type == Enrichment.CAZY:
            desc = None

        if self.annotation_type == Enrichment.TIGRFAM:
            desc = self.tigrfamdescription

        if self.annotation_type == Enrichment.PFAM:
            desc = self.pfam2description
            
        if self.annotation_type == Enrichment.EC:
            desc = self.ec2description

        for line in output_lines:
            annotation = line[0]
            
            if desc:
            
                if annotation in desc:
                    line.append(desc[annotation])
            
                else:
                    line.append("NA")
            
            else:
                line.append("NA")
        
        return output_lines

    def do(self, group_dict):
        results = list()

        for combination in combinations(group_dict, 2):
            enrichment_test, overrepresentation_test = self.test_chooser( [group_dict[member] for member in combination] )
            prefix = '_vs_'.join(combination).replace(' ', '_')
            logging.info('Comparing gene frequency among groups: %s' % ', '.join(combination))
            
            if enrichment_test == stats.fisher_exact:
                logging.info('Testing gene enrichment using Fisher\'s exact test')    
                gene_count = self.gene_frequencies(*combination)
                output_lines = self.pool.map(gene_fisher_calc, gene_count)
            
                for idx, corrected_pval in enumerate(self.corrected_pvals(output_lines)):
                    output_lines[idx].append(str(corrected_pval))

                output_lines = self.add_descriptions(output_lines)
                output_lines = self.FISHER_HEADER + output_lines
                results.append([output_lines, prefix +'_'+ self.GENE_FISHER_OUTPUT])
            
            elif enrichment_test == self.PA:
                logging.info('enrichment statistics not possible with only one genome to compare')
                logging.info('See prevalence matrix for unique genes in groups')

            logging.info('Comparing gene over-representation among genomes')
            if overrepresentation_test == stats.mannwhitneyu:
                logging.info('Testing over-representation using Mann-Whitney U test')
                gene_count = self.gene_frequencies(*combination, True)
                output_lines = self.pool.map(mannwhitneyu_calc, gene_count)

                for idx, corrected_pval in enumerate(self.corrected_pvals(output_lines)):
                    output_lines[idx].append(str(corrected_pval))
                
                output_lines = self.add_descriptions(output_lines)
                output_lines = self.MANNWHITNEYU_HEADER + output_lines 
                results.append([output_lines, prefix +'_'+ self.GVG_OUTPUT])

            elif overrepresentation_test == stats.norm.cdf:
                logging.info('Testing over-representation using Z score test')
                gene_count = self.gene_frequencies(*combination, True)
                output_lines = self.pool.map(zscore_calc, gene_count)
                output_lines = [x for x in output_lines if x]
                
                for idx, corrected_pval in enumerate(self.corrected_pvals(output_lines)):
                    output_lines[idx].append(str(corrected_pval))

                output_lines = self.add_descriptions(output_lines)
                output_lines = self.ZSCORE_HEADER + output_lines
                results.append([output_lines, prefix +'_'+ self.IVG_OUTPUT])

        return results







#!/usr/bin/env python3
###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

__author__      = "Joel Boyd"
__copyright__   = "Copyright 2017"
__credits__     = ["Joel Boyd"]
__license__     = "GPL3"
__version__     = "0.0.7"
__maintainer__  = "Joel Boyd"
__email__       = "joel.boyd near uq.net.au"
__status__      = "Development"
 
###############################################################################

class NetworkTraverser:
    
    to_omit = set(["C00828",  # Menaquinone
                   "C00534",  # Pyridoxamine
                   "C00006",  # NADP+
                   "C00003",  # NAD+
                   "C00002",  # ATP
                   "C00314",  # Pyridoxine
                   "C00864",  # Pantothenate
                   "C00504",  # Folate
                   "C00032",  # Heme
                   "C05443",  # Vitamin D3
                   "C00253",  # Nicotinate
                   "C00250",  # Pyridoxal
                   "C11378",  # Ubiquinone-10
                   "C05777",  # Coenzyme F430
                   "C00072",  # Ascorbate
                   "C00378",  # Thiamine
                   "C00101",  # Tetrahydrofolate
                   "C00029",  # UDP-glucose
                   "C00068",  # Thiamin diphosphate
                   "C00061",  # FMN
                   "C00063",  # CTP
                   "C05776",  # Vitamin B12
                   "C00113",  # PQQ
                   "C18237",  # Molybdoenzyme molybdenum cofactor
                   "C00051",  # Glutathione
                   "C00010",  # CoA
                   "C00016",  # FAD
                   "C00018",  # Pyridoxal phosphate
                   "C00019",  # S-Adenosyl-L-methionine
                   "C00153",  # Nicotinamide
                   "C04628",  # Coenzyme B
                   "C00862",  # Methanofuran
                   "C15672",  # Heme O
                   "C15670",  # Heme A
                   "C02059",  # Phylloquinone
                   "C03576",  # Coenzyme M
                   "C05441",  # Vitamin D2
                   "C00272",  # Tetrahydrobiopterin
                   "C02477",  # alpha-Tocopherol
                   "C00473",  # Retinol
                   "C00120",  # Biotin
                   "C00725",  # Lipoate
                   "C00053",  # 3'-Phosphoadenylyl sulfate
                   "C00194",  # Cobamide coenzyme
                   "C00255",  # Riboflavin
                   'C00001',  # H2O
                   'C00008',  # ADP
                   'C00013',  # Diphosphate
                   'C00004',  # NADH
                   'C00005',  # NADPH
                   'C00080',  # H+
                   'C00009',  # Orthophosphate
                   'C00008',  # ADP
                   'C00004',  # NADH
                   'C00020',  # AMP
                   'C00007',  # Oxygen
                   'C00015']) # UDP

    @staticmethod
    def dijkstra_traverse(from_node, to_node):
        pass

    @staticmethod  
    def shortest_bfs_path(r2rpair_graph, c2r_graph, from_node, to_node):
    # Modified from:
    # http://eddmann.com/posts/depth-first-search-and-breadth-first-search-in-python/
        def bfs_traverse(r2rpair_graph, c2r_graph, from_node, to_node):
            queue = [(from_node, [from_node])]
            while queue:
                (vertex, path) = queue.pop(0)     
                for reaction in c2r_graph[vertex]:
                    for rpair in r2rpair_graph[reaction]:
                        if vertex in rpair:
                            compound = [x for x in rpair.split('_') 
                                          if x != vertex].pop()
                            new_path = path + [reaction, compound]
                            if compound==to_node:
                                yield path + [reaction, compound]
                            else:
                                queue.append((compound, new_path))

        try:
            return next(bfs_traverse(r2rpair_graph, c2r_graph, from_node, 
                                     to_node))
        except StopIteration:
            return None


#!/usr/bin/env python3
###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

__author__      = "Joel Boyd"
__copyright__   = "Copyright 2018"
__credits__     = ["Joel Boyd"]
__license__     = "GPL3"
__version__     = "0.0.7"
__maintainer__  = "Joel Boyd"
__email__       = "joel.boyd near uq.net.au"
__status__      = "Development"

###############################################################################
# Imports
import logging
import pickle
import os
import numpy as np
from sklearn.ensemble import RandomForestRegressor, RandomForestClassifier
# Local
from enrichm.generate import GenerateModel
################################################################################


class Predict():

	def __init__(self):
		self.PREDICTIONS_OUTPUT_PATH = 'predictions.tsv'

		self.LABELS_DICT = "labels_dict.pickle"
		self.RF_MODEL = "rf_model.pickle"
		self.ATTRIBUTE_LIST = "attribute_list.txt"

	def _write_predictions(self, output_lines, output_directory):
		'''		
		Inputs
		------
		
		Outputs
		-------
		
		'''
		output_path = os.path.join(output_directory, self.PREDICTIONS_OUTPUT_PATH)
		logging.info('Writing predictions to output %s' % (output_path))
		with open(output_path, 'wb') as out_io:
			out_io.write(str.encode('\t'.join(["Sample", "Prediction", "Probability"]) + '\n'))
			for line in output_lines:
				out_io.write(str.encode(line + '\n'))

	def _make_predictions(self, model, sample_list, content_list, attribute_dictionary):
		'''		
		Inputs
		------
		
		Outputs
		-------
		
		'''
		
		sample_list 	= np.array(sample_list)
		content_list 	= np.array(content_list)
		predictions 	= model.predict(content_list)
		probabilities 	= model.predict_proba(content_list)

		output_list = []

		for sample, prediction, probability in zip(sample_list, predictions, probabilities):
			max_prob = str(round(max(list(probability)), 2))
			prediction = str(attribute_dictionary[prediction])
			output_list.append('\t'.join([sample, prediction, max_prob]))
		
		return output_list

	def parse_input_model_directory(self, forester_model_directory):

		output_dictionary = {
			self.LABELS_DICT: None,
			self.RF_MODEL: None,
			self.ATTRIBUTE_LIST:None
		}

		contents = os.listdir(forester_model_directory)

		for content in contents:
			content_path = os.path.join(forester_model_directory, content)
			if content in output_dictionary:
				if content.endswith("pickle"):
					output_dictionary[content] = pickle.load(open(content_path, 'rb'))
				else:
					output_dictionary[content] = [x.strip() for x in open(content_path)]

		if None in list(output_dictionary.values()):
			raise Exception("Malformatted forester model directory: %s" % (forester_model_directory))

		return output_dictionary


	def do(self, forester_model_directory, input_matrix_path, output_directory):
		'''		
		Inputs
		------
		
		Outputs
		-------
		
		'''
		forester_model = self.parse_input_model_directory(forester_model_directory)		

		logging.info('Parsing input')
		gm = GenerateModel()
		logging.info('Loading model: %s' % (self.RF_MODEL))

		logging.info('Parsing data')
		features, _ \
			= gm.parse_input_matrix(input_matrix_path)
		
		sample_list = list()
		content_list = list()
		for sample, content in features.items():
			sample_list.append(sample)
			sample_content = []
			for attribute in forester_model['attribute_list.txt']:
				if attribute in content:
					sample_content.append(content[attribute])
				else:	
					sample_content.append('0')
					
			content_list.append(sample_content)

		logging.info('Making predictions')
		output_lines = self._make_predictions(forester_model[self.RF_MODEL],
											  sample_list,
										  	  content_list,
										  	  forester_model[self.LABELS_DICT])
		self._write_predictions(output_lines, output_directory)



#!/usr/bin/env python3
###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

__author__      = "Joel Boyd"
__copyright__   = "Copyright 2017"
__credits__     = ["Joel Boyd"]
__license__     = "GPL3"
__version__     = "0.0.7"
__maintainer__  = "Joel Boyd"
__email__       = "joel.boyd near uq.net.au"
__status__      = "Development"

###############################################################################
# System imports
import logging
import os
from collections import Counter
# Local imports
from enrichm.parse_annotate import ParseAnnotate
from enrichm.databases import Databases
from enrichm.module_description_parser import ModuleDescription

###############################################################################

class Connect(object):

	def __init__(self):
		d = Databases()
		
		self.m2def = d.m2def
		self.m2c = d.m2c
		self.c2m = dict()

		for module, compounds in self.m2c.items():	
			substrates = compounds[0]
			for substrate in substrates:
				if substrate in self.c2m:
					self.c2m[substrate].append(module)
				else:
					self.c2m[substrate] = [module]

		self.c = d.c
		self.m = d.m
		self.signature_modules = d.signature_modules
		self.output_file = 'linkages.tsv'
	
	def _update_with_custom_modules(self, custom_modules):
		custom_modules_dict = {line.split('\t')[0]:line.strip().split('\t')[1]
								for line in open(custom_modules)}
		self.m2def.update(custom_modules_dict)
		
		for key in custom_modules_dict.keys():
			self.m[key] = 'Custom'

	def _parse_genome_and_annotation_file_matrix(self, genome_and_annotation_file):
		genome_and_annotation_file_io = open(genome_and_annotation_file)
		headers = genome_and_annotation_file_io.readline().strip().split('\t')[1:]
		genome_to_annotation_sets = {genome_name:set() for genome_name in headers}

		for line in genome_and_annotation_file_io:
			sline = line.strip().split('\t')
			annotation, entries = sline[0], sline[1:]

			for genome_name, entry in zip(headers, entries):

				if float(entry) > 0.0:
					genome_to_annotation_sets[genome_name].add(annotation)

		return genome_to_annotation_sets

	def _write(self, lines, output_path):
		with open(output_path, 'wb') as out_io:

			for l in lines:
				out_io.write(('\t'.join(l) + '\n').encode())

	def do(self, annotate_output, metadata, custom_modules, cutoff, output_directory):
		pa = ParseAnnotate(annotate_output, 1)
		annotations = self._parse_genome_and_annotation_file_matrix(pa.ko)

		if custom_modules:
			self._update_with_custom_modules(custom_modules)

		genomes_list = list(annotations.keys())
		genomes_modules = dict()
		module_to_genome = dict()
		candidate_linkages = [["Genome 1",
					   		   "Genome 2",
					   		   "Compound",
					   		   "Compound description",
					   		   "Genome 1 module (producer)",
					   		   "Genome 2 module (consumer)"]]

		for module in self.m2def:

			for genome_name in genomes_list:
				kos = annotations[genome_name]

				if module not in self.signature_modules:
					path = ModuleDescription(self.m2def[module])
					num_all = path.num_steps()
					num_covered, _, _, _ = path.num_covered_steps(kos)
					perc_covered = num_covered / float(num_all)

					if perc_covered>= float(cutoff):
						if module in module_to_genome:
							module_to_genome[module].add(genome_name)
						else:
							module_to_genome[module] = set([genome_name])

						if genome_name in genomes_modules:
							genomes_modules[genome_name].add(module)
						else:
							genomes_modules[genome_name] = set([module])

		for genome_name, modules in genomes_modules.items():

			for module in modules:
				
				if module in self.m2c:
					produced_compounds = self.m2c[module][1]

					for compound in produced_compounds:

						if compound in self.c2m:
							possible_modules 	= self.c2m[compound]
							modules_not_covered = list()

							for possible_module in possible_modules:

								if possible_module not in modules:
									modules_not_covered.append(possible_module)
							
							for nc_module in modules_not_covered:
								
								if nc_module in module_to_genome:
									candidate_genomes = module_to_genome[nc_module]

									for genome in candidate_genomes:
										
										if genome != genome_name:
											encoded_modules = genomes_modules[genome]
											
											if module not in encoded_modules:
												candidate_linkages.append([genome_name,
																		   genome,
																		   compound,
																		   self.c[compound],
																		   module,
																		   nc_module])

		self._write(candidate_linkages, os.path.join(output_directory, self.output_file))




#!/usr/bin/env python3
###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

from enrichm.databases import Databases
from enrichm.sequence_io import SequenceIO
import logging
import os
__author__ = "Joel Boyd"
__copyright__ = "Copyright 2017"
__credits__ = ["Joel Boyd"]
__license__ = "GPL3"
__version__ = "0.0.7"
__maintainer__ = "Joel Boyd"
__email__ = "joel.boyd near uq.net.au"
__status__ = "Development"

###############################################################################

class Genome:
	'''
	A genome object which collects all the attirbutes of an imput genome,
	including protein sequences and their annotations
	'''
	def __init__(self, light, path, nucl):
		seqio = SequenceIO()
		self.clusters = set()
		self.orthologs = set()
		self.protein_ordered_dict = dict()
		self.sequences = dict()
		self.cluster_dict = dict()
		self.ortholog_dict = dict()
		self.path = path
		self.name = os.path.split(os.path.splitext(path)[0])[1]
		
		if light == False:
		
			if nucl is not None:
				self.nucl 	= nucl
				self.length = 0
				gc_list 	= 0.0

				for description, sequence in seqio.each(open(nucl)):
					self.length += len(str(sequence))
					gc_list 	+= (str(sequence).count('G') + str(sequence).count('C'))
				
				self.gc = round((gc_list/float(self.length))*100, 2)
		
			for protein_count, (description, sequence) in enumerate(seqio.each(open(path))):
				name = description.partition(' ')[0]
				sequence = Sequence(description, sequence)
				self.sequences[name] = sequence
				self.protein_ordered_dict[protein_count] = name
		
		else:
			for protein_count, (description, _) in enumerate(seqio.each(open(path))):
				name = description.partition(' ')[0]
				sequence = Sequence(description)
				self.sequences[name] = sequence
				self.protein_ordered_dict[protein_count] = name

	def add(self, annotations, evalue_cutoff, bitscore_cutoff, 
		    percent_aln_query_cutoff, percent_aln_reference_cutoff, 
		    annotation_type):
		'''
		Adds a series of annotations to the proteins within a genome.

		Parameters
		----------
		annotations						- String. Path to file containing either blast or
										  domtblout hmmsearch, or mmseqs cluster results. 
		evalue_cutoff					- Float. E-value threshold for annotations.
		bitscore_cutoff					- Float. Bit score threshold for annotations.
		percent_aln_query_cutoff		- Float. Threshold for the percent of the query 
										  that must be aligned to consider the annotation.
		percent_aln_reference_cutoff	- Float. Threshold for the percent of the reference
										  that must be aligned to consider the annotation.
		annotation_type					- String. Either 'KO', 'TIGRFAM', 'PFAM', 	
										  'HYPOTHETICAL' or 'COG'
		'''
		# Load up annotation parser, and tell it what annotation type to expect
		ap = AnnotationParser(annotation_type)

		# If annotation type is a hmmsearch result
		if(annotation_type == AnnotationParser.PFAM or
		   annotation_type == AnnotationParser.TIGRFAM or
		   annotation_type == AnnotationParser.CAZY or
		   annotation_type == AnnotationParser.KO_HMM):
			# Set up an iterator to produce the results
			logging.debug("    - Parsing hmmsearch chunk")

			if(annotation_type == AnnotationParser.PFAM or
			   annotation_type == AnnotationParser.KO_HMM):
				percent_aln_query_cutoff = 0.0
				percent_aln_reference_cutoff = 0.0

			iterator = ap.from_hmmsearch_results(annotations, evalue_cutoff,
												 bitscore_cutoff, percent_aln_query_cutoff, 
												 percent_aln_reference_cutoff,
                                       			 (True if annotation_type == AnnotationParser.KO_HMM else False))
			
			if annotation_type == AnnotationParser.PFAM:
				self.pfam_dict = dict()
				refdict = self.pfam_dict

			elif annotation_type == AnnotationParser.KO_HMM:
				self.ko_dict = dict()
				refdict = self.ko_dict
			
			elif annotation_type == AnnotationParser.TIGRFAM:
				self.tigrfam_dict = dict()
				refdict = self.tigrfam_dict
			
			elif annotation_type == AnnotationParser.CAZY:
				self.cazy_dict = dict()
				refdict = self.cazy_dict

		# If annotation type is a blast result
		
		elif(annotation_type == AnnotationParser.KO or
			 annotation_type == AnnotationParser.EC):
			# Set up an iterator to produce the results
			logging.debug("    - Parsing blast chunk")
			iterator = ap.from_blast_results(annotations, evalue_cutoff, 
											 bitscore_cutoff, percent_aln_query_cutoff)

			if annotation_type == AnnotationParser.KO:
				self.ko_dict = dict()
				refdict = self.ko_dict

			elif annotation_type == AnnotationParser.EC:
				self.ec_dict = dict()
				refdict = self.ec_dict
		
		for seqname, annotations, evalue, annotation_range in iterator:
			self.sequences[seqname].add(annotations, evalue, annotation_range, annotation_type)
			for annotation in annotations:

				if annotation in refdict:
					refdict[annotation].append(seqname)
				
				else:
					refdict[annotation]=[seqname]
			
	def count(self, annotation, type):
		'''
		
		Parameters
		----------
		annotation - String. An annotation ID to return a frequency for
		
		Output
		------
		The number of times this annotation was encountered in the genome
		'''

		if type==AnnotationParser.HYPOTHETICAL:
			reference_dict = self.cluster_dict
		
		elif type==AnnotationParser.KO:
			reference_dict = self.ko_dict
		
		elif type==AnnotationParser.EC:
			reference_dict = self.ec_dict
		
		elif type==AnnotationParser.PFAM:
			reference_dict = self.pfam_dict
		
		elif type==AnnotationParser.TIGRFAM:
			reference_dict = self.tigrfam_dict
		
		if annotation in reference_dict:
			return len(reference_dict[annotation])
		
		else:
			return 0

	def ordered_sequences(self):
		'''
		Iterator that yields that all protein coding Sequence objects in a genome in order.
		'''
		for sequence_id in sorted(self.protein_ordered_dict.keys()):
			
			yield self.sequences[self.protein_ordered_dict[sequence_id]]

	def add_cluster(self, sequence_id, cluster_id):
		'''
		Add hypothetical clusters to genome sequences
		
		Inputs
		------
		cluster_list - array. list where each entry is a list
					   of n = 2: sequence_name, cluster_name
		'''
	
		if cluster_id in self.cluster_dict:
			self.cluster_dict[cluster_id].append(sequence_id)
		
		else:
			self.cluster_dict[cluster_id] = [sequence_id]
		
		annotation = Annotation(cluster_id, 0, [-1,0], AnnotationParser.HYPOTHETICAL)
		
		self.sequences[sequence_id].cluster = cluster_id
		self.sequences[sequence_id].annotations.append(annotation)
		
		self.clusters.add(cluster_id)
	
	def add_ortholog(self, sequence_id, ortholog_id):
		'''
		Add hypothetical ortholog to genome sequences
		
		Inputs
		------
		cluster_list - array. list where each entry is a list
					   of n = 2: sequence_name, cluster_name
		'''
	
		if ortholog_id in self.ortholog_dict:
			self.ortholog_dict[ortholog_id].append(sequence_id)
		
		else:
			self.ortholog_dict[ortholog_id] = [sequence_id]
		
		annotation = Annotation(ortholog_id, 0, [-1,0], AnnotationParser.ORTHOLOG)
		
		self.sequences[sequence_id].ortholog = ortholog_id
		self.sequences[sequence_id].annotations.append(annotation)
		self.orthologs.add(ortholog_id)

class Sequence(Genome):
	'''
	Sequence object which collects all attributes of a sequence including its length,
	and annotations. Can compare current annotation with new annotaitons.
	'''
	def __init__(self, description, sequence=None):
		self.annotations = list()
		line_split = description.split(' # ')
		
		if sequence:
			self.seq = str(sequence)
			self.length = int(len(sequence))	

		if len(line_split)==5:
			self.seqname, self.startpos, self.finishpos, self.direction, stats \
								= line_split
			self.prod_id, self.partial, self.starttype, self.rbs_motif, self.rbs_spacer,  self.gc \
								= [x.split('=')[1] for x in stats.split(';')]
		
		elif len(line_split)==1:
			self.seqname = description
			
		else:
			raise Exception("Malformatted sequence file!")

	def all_annotations(self):
		'''
		Returns a list of all annotations assigned to this sequence
		'''
		result = []
		
		for annotation in self.annotations:
			result.append(annotation.annotation)
		
		return result

	def seqdict(self):
		'''
		Output
		------
		A dictionary where each entry is a position in the sequence,
		and values are the annotation at that position. This is important
		in particular for pfam annotations as a protein can have >1 
		domains.
		'''
		seq_dict = {x:None for x in range(self.length)}
		
		for annotation in self.annotations:
		
			for position in annotation.region:
				seq_dict[position] = annotation.annotation
		
		return seq_dict

	def what(self, query_region):
		'''
		Return annotations assigned to a list of positions within a sequence.

		Parameters
		----------
		region 		- list. List of integers specifying the positions in the 
					  sequence to return annotations for
		Outputs
		-------
		Returns a list of equal length to region, containing the annotation of
		each position.
		'''
		result = list()

		# Build reference dictionary for sequence
		seq_dict = self.seqdict()

		# Find annotation for each position in the sequence
		for position in query_region:
			result.append(seq_dict[position])
		
		return result

	def add(self, annotations, evalue, region, annotation_type):
		'''
		Return annotations assigned to a list of positions within a sequence.

		Parameters
		----------
		annotation 	- string. Annotation to assign to the sequence region
		evalue 		- float. Evalue awarded for the given annotation
		region		- list. List of integers specifying the positions in the 
					  sequence to annotate
		'''

		new_annotations = [Annotation(annotation, evalue, region, annotation_type) for annotation in annotations]
		annotation_list = [annot for annot in self.annotations if annot.type == new_annotations[0].type]
		
		if len(annotation_list) > 0:
			
			to_remove 	= list()
			to_check 	= annotation_list

			for new_annotation in new_annotations:
				overlap 	= [previous_annotation for previous_annotation in to_check
							   if len(previous_annotation.region.intersection(new_annotation.region)) > 0]
				
				if annotation_type==AnnotationParser.PFAM:
					self.annotations.append(new_annotation)
				
				else:
					
					if len(overlap)>0:
						
						for overlapping_previous_annotation in overlap:
							
							if new_annotation.compare(overlapping_previous_annotation):
								to_remove.append(overlapping_previous_annotation)

						if len(to_remove)>0:
							self.annotations = [annotation for annotation in self.annotations 
												if annotation not in to_remove]
							self.annotations.append(new_annotation)
					
					else:
							self.annotations.append(new_annotation)

		else:
			
			for new_annotation in new_annotations:
				self.annotations.append(new_annotation)
	
class Annotation(Sequence):
	'''
	Annotation object that collects all attributes associated with a given
	annotation, like where it is in the sequence, its e-value and bit score,
	and what type of annotation it is.
	'''
	def __init__(self, annotation, evalue, region, annotation_type):
		self.annotation = annotation
		self.evalue 	= float(evalue)
		self.region		= set(region)
		self.type   	= annotation_type

	def compare(self, other_annotation):
		'''
		Compares evalue of current annotation with the evaulue of another annotation object. 			

		Parameters
		----------
		other_annotation: Annotation object.
		
		Output
		------
		Returns True if self is the better annotation or False if it isn't
		'''
		if self.evalue < other_annotation.evalue:
			return True
		
		else:
			return False

class AnnotationParser:
	'''
	Annotation parser class contains fucntions to parse hmmsearch domtblouts and blast results 
	currently for: KO, PFAM and TIGRFAM. COG to come	
	'''
	KO      		= 'KO_IDS.txt'
	KO_HMM 			= 'KO_IDS.txt'
	EC				= 'EC_IDS.txt'
	PFAM    		= 'PFAM_IDS.txt'
	TIGRFAM 		= 'TIGRFAM_IDS.txt'
	CAZY      		= 'CAZY_IDS.txt'
	HYPOTHETICAL 	= 'HYPOTHETICAL.txt'
	ORTHOLOG 		= 'ORTHOLOG.txt'

	def __init__(self, annotation_type):        
		pass
	
	def from_blast_results(self,
						   blast_output,
						   evalue_cutoff,
						   bitscore_cutoff, 
						   percent_id_cutoff):
		'''
		Parse blast output in tab format.	

		Parameters
		----------
		blast_output 		- String. Path to blast output file containing results.
							  Must be in blast output format 6
		evalue_cutoff		- Float. E-value threshold for annotations.
		bitscore_cutoff		- Float. Bit score threshold for annotations.
		percent_id_cutoff 	- Float. Percent identity threshold for annotations.
		
		Yields
		------
		A sequence name, annotation, E-value and region hit for every annotation result in 
		blast_output that pass a series of specified cutoffs
		'''

		for sline in blast_output:
			# Parse out important information from each line in blast output
			evalue   = sline[10]
			bit      = sline[11]
			perc_id  = sline[2]
			seq_list = [int(x) for x in sline[6:8]]

			# If the annotation passes the specified cutoffs
			if(float(evalue) <= evalue_cutoff and
				float(bit) >= bitscore_cutoff and
				float(perc_id) >= percent_id_cutoff):
					seqname = sline[0].split('~')[1]
					annotation = sline[1].split('~')[1].split('+')
					
					yield seqname, annotation, evalue, range(min(seq_list), max(seq_list))

	def from_hmmsearch_results(self,
							   hmmsearch_output_path,
							   evalue_cutoff,
							   bitscore_cutoff, 
    						   percent_aln_query_cutoff,
    						   percent_aln_reference_cutoff,
							   acc = False):
		'''
		Parse input hmmsearch file

		Parameters 
		----------
		hmmsearch_output_path           - String. Path to domtblout file containing 
										  hmmsearch results.
		evalue_cutoff                   - Float. E-value threshold for annotations.
		bitscore_cutoff                 - Float. Bit score threshold for annotations.
		percent_aln_query_cutoff        - Float. Threshold for the percent of the query 
										  that must be aligned to consider the annotation.
		percent_aln_reference_cutoff    - Float. Threshold for the percent of the reference 	
										  that must be aligned to consider the annotation.
		Yields
		------
		A sequence name, accession, E-value and region hit for every annottation result in 
		blast_output_path that pass a set of cutoffs
		'''
		
		# Filling in column
		for line in open(hmmsearch_output_path):
			
			# Skip headers
			if line.startswith('#'): continue
				
			# Parse HMMsearch line. '_'s represent unimportant entries. Line
			# is trimmed using [:22] to remove sequence description
			seqname, _, tlen, ko_hmm, accession, qlen, _, score, \
			_, _, _, _, i_evalue, _, _, _, \
			_, seq_from, seq_to, _, _, _ = line.strip().split()[:22]				

			# Determine sequence and HMM spans
			seq_list = [int(seq_from), int(seq_to)]

			# Calculate percent of the query and reference aligned to each-other. 
			perc_seq_aln = (max(seq_list)-min(seq_list))/float(tlen)
			perc_hmm_aln = (max(seq_list)-min(seq_list))/float(qlen)
			
			if acc:
				accession = ko_hmm

			# If the annotation passes the specified cutoffs
			if(float(i_evalue)<=evalue_cutoff and
				float(score)>=bitscore_cutoff and
				perc_seq_aln>=percent_aln_query_cutoff and
				perc_hmm_aln>=percent_aln_reference_cutoff):
				yield seqname, [accession], i_evalue, range(min(seq_list), max(seq_list))



#!/usr/bin/env python3
###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import logging
import os
import itertools
from enrichm.network_builder import NetworkBuilder
from enrichm.parser import Parser
from enrichm.databases import Databases
from enrichm.kegg_matrix import KeggMatrix

__author__ = "Joel Boyd"
__copyright__ = "Copyright 2017"
__credits__ = ["Joel Boyd"]
__license__ = "GPL3"
__version__ = "0.0.7"
__maintainer__ = "Joel Boyd"
__email__ = "joel.boyd near uq.net.au"
__status__ = "Development"

###############################################################################

class NetworkAnalyser:
    
    MATRIX          = 'matrix'
    NETWORK         = 'network'
    EXPLORE         = 'explore'
    DEGRADE         = 'degrade'
    PATHWAY         = 'pathway'
    ANNOTATE        = 'annotate'
    ENRICHMENT      = 'enrichment'
    MODULE_AB       = 'module_ab'
    TRAVERSE        = 'traverse'

    NETWORK_OUTPUT_FILE  = 'network.tsv'
    METADATA_OUTPUT_FILE = 'metadata.tsv'    
    TRAVERSE_OUTPUT_FILE = 'traverse.tsv'    

    def __init__(self, metadata):
        self.metadata = dict()
        
        for line in open(metadata):
            
            if line.startswith('#'): 
                continue
            
            split_line = line.strip().split('\t')
            
            if len(split_line) == 1:
                raise Exception("Only one column detected in metadata file, please check that your file is tab separated")
            else:
                sample_id, group = split_line 

            if group in self.metadata:
                self.metadata[group].append(sample_id)
            else:
                self.metadata[group] = [sample_id]

    def _write_results(self, output_path, output_lines):
        '''
        Parameters
        ----------
        output_path: string
            Path to non-existent file to write output lines to
        output_lines: list
            list containing lines to write to output path
        '''
        logging.info('Writing results to file: %s' % output_path)

        with open(output_path, 'w') as output_path_io: 
            output_path_io.write('\n'.join(output_lines))
            output_path_io.flush()
    
    def _average(self, d):
            
        for sample_group, group_dict in d.items():

            for group, reaction_dict in group_dict.items():
            
                for reaction, value in reaction_dict.items():
                    d[sample_group][group][reaction] = sum(value) / len(value)
        
        return d

    def normalise_by_abundance(self, sample_abundance_dict, sample_metadata, reaction_abundance_dict, metadata):

        reversed_metadata = dict()
        for key, items in metadata.items():

            for item in items:
                reversed_metadata[item] = key
        
        averaged_sample_abundance = dict()
        
        for group, samples in sample_metadata.items():
            averaged_sample_abundance[group] = dict()
            dicts = [sample_abundance_dict[x] for x in samples]
            genomes = set(list(itertools.chain(*[list(x.keys()) for x in dicts])))
            
            for genome in genomes:
                abundances = [d[genome] for d in dicts]
                average = sum(abundances) / len(abundances)
                averaged_sample_abundance[group][genome] = average # Median might be better?
        new_dict = {x:dict() for x in list(averaged_sample_abundance.keys())}

        for sample_group, genome_abundances in averaged_sample_abundance.items():
            
            for genome, genome_abundance in genome_abundances.items():

                if genome in reversed_metadata:

                    if genome in reaction_abundance_dict[sample_group]:

                        for reaction in list(reaction_abundance_dict[sample_group][genome].keys()):
                            
                            if reaction in reaction_abundance_dict[sample_group][genome]:
                                normalised_value = reaction_abundance_dict[sample_group][genome][reaction]*genome_abundance             

                            else:
                                normalised_value = 0.0
                            group = reversed_metadata[genome]

                            if group not in new_dict[sample_group]:
                                new_dict[sample_group][group] = dict()
                            
                            if reaction not in new_dict[sample_group][group]:
                                new_dict[sample_group][group][reaction] = [0.0]

                            new_dict[sample_group][group][reaction].append( normalised_value )
        
        new_dict = self._average(new_dict) # taking averages here again, might be better accumulated?

        return new_dict

    def _parse_enrichment_output(self, enrichment_output):
        fisher_results = dict()
        
        for file in os.listdir(enrichment_output):

            if file.endswith("fisher.tsv"):
                file = os.path.join(enrichment_output, file)
                file_io = open(file)
                file_io.readline()

                for line in file_io:
                    split_line = line.strip().split('\t')
                     
                    if len(fisher_results) == 0:
                        fisher_results[split_line[1]] = list()
                        fisher_results[split_line[2]] = list()

                    if float(split_line[-2])<0.05:
                        g1_t = float(split_line[3])
                        g1_f = float(split_line[4])
                        g2_t = float(split_line[5])
                        g2_f = float(split_line[6])

                        if g1_t == 0:
                            fisher_results[split_line[2]].append( split_line[0] )

                        elif g2_t == 0:
                            fisher_results[split_line[1]].append( split_line[0] )

                        elif ( ((g1_t/(g1_t+g1_f))) / ((g2_t/(g2_t+g2_f))) )>1:
                            fisher_results[split_line[1]].append( split_line[0] )

                        else:
                            fisher_results[split_line[2]].append( split_line[0] )
                
        if len(fisher_results.keys())>0:
            return fisher_results
        
        else:
            raise Exception("Malformatted enrichment output: %s" % enrichment_output)
    
    def average_tpm_by_sample(self, tpm_results, sample_metadata, group_metadata):
        output_dict = dict()
        tpm_dict, annotations, genomes = tpm_results

        for group, samples in sample_metadata.items():
            output_dict[group] = dict()

            for sample in samples:
            
                for annotation in annotations:
                                
                    if str.encode(sample) in tpm_dict:
                        
                        for genome in genomes:

                            if genome not in output_dict[group]:
                                output_dict[group][genome] = dict()
        
                            if annotation not in output_dict[group][genome]:
                                output_dict[group][genome][annotation] = list()
                            if genome in tpm_dict[str.encode(sample)]:

                                if annotation in tpm_dict[str.encode(sample)][genome]:
                                    output_dict[group][genome][annotation].append(tpm_dict[str.encode(sample)][genome][annotation])
                                
                                else:    
                                    output_dict[group][genome][annotation].append(0.0)
                            
                            else:      
                                output_dict[group][genome][annotation].append(0.0)
            
            for genome, values in output_dict[group].items():

                for annotation in values:
                    output_dict[group][genome][annotation] = sum(output_dict[group][genome][annotation])/len(output_dict[group][genome][annotation])
        
        new_output_dict = dict()
        reactions = list(Databases().r.keys())
        for key, item in output_dict.items():
            new_output_dict[key] = dict()

            for group, members in group_metadata.items():
                new_output_dict[key][group] = dict()

                for reaction in reactions:
                    to_average = list()
                    for member in members:
                        if member in item:
                            if str.encode(reaction) in item[member]:
                                to_average.append(item[member][str.encode(reaction)])
                            else:
                                to_average.append(0.0)
                    average_value = sum(to_average) / len(to_average)
                    new_output_dict[key][group][reaction] = average_value

        return new_output_dict

    def do(self,
            matrix,
            tpm_values,
            abundance,
            abundance_metadata,
            metabolome,
            enrichment_output,
            depth,
            filter,
            limit,
            queries,
            starting_compounds,
            steps,
            number_of_queries,
            output_directory):
        '''
        Parameters
        ----------
        matrix
        tpm_values
        abundance
        abundance_metadata
        metabolome
        enrichment_output
        depth
        filter
        limit
        queries
        starting_compounds
        steps
        number_of_queries
        output_directory
        '''

        km = KeggMatrix(matrix)
        nb = NetworkBuilder(self.metadata.keys())

        # Read in fisher results
        if enrichment_output:
            fisher_results = self._parse_enrichment_output(enrichment_output)
        else:
            fisher_results = None    

        # Read in genome abundance
        if abundance:
            sample_abundance = km._parse_matrix(abundance)
            sample_metadata = list(km._parse_matrix(abundance_metadata).values())[0]
            d = dict()
            
            for key, item in sample_metadata.items():
                
                if item not in d:
                    d[item] = list()
                
                d[item].append(key)
            
            sample_metadata = d
        
        else:
            sample_abundance = {'MOCK': {x:1 for x in list(km.reaction_matrix.keys())} }
            sample_metadata = {"a": ['MOCK']}
            km.reaction_matrix = {"a": km.reaction_matrix}
        
        # Read in expression (TPM) values
        if tpm_values:
            logging.info("Parsing detectM TPM values")
            normalised_abundances \
                    = self.average_tpm_by_sample(Parser.parse_tpm_values(tpm_values), sample_metadata, self.metadata)

        else:
            normalised_abundances \
                = self.normalise_by_abundance(sample_abundance,
                                              sample_metadata,
                                              km.reaction_matrix,
                                              self.metadata)
        if metabolome:
            abundances_metabolome = km._parse_matrix(metabolome)
        
        else:
            abundances_metabolome = None

        logging.info('Generating pathway network')

        network_lines, node_metadata = \
                        nb.pathway_matrix(normalised_abundances, 
                                          abundances_metabolome,
                                           fisher_results,
                                           limit,
                                           filter)

        self._write_results(os.path.join(output_directory, self.NETWORK_OUTPUT_FILE), network_lines)
        self._write_results(os.path.join(output_directory, self.METADATA_OUTPUT_FILE), node_metadata)



#!/usr/bin/env python3
###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

__author__      = "Joel Boyd"
__copyright__   = "Copyright 2017"
__credits__     = ["Joel Boyd"]
__license__     = "GPL3"
__version__     = "0.0.7"
__maintainer__  = "Joel Boyd"
__email__       = "joel.boyd near uq.net.au"
__status__      = "Development"
 
###############################################################################
# Imports
import os
import logging
import pickle
# Local
from enrichm.data import Data
###############################################################################


class Databases:
	
	if os.path.isfile(os.path.join(Data.DATABASE_DIR, 'VERSION')):
		DB_VERSION			= open(os.path.join(Data.DATABASE_DIR, 'VERSION')).readline().strip().replace('.tar.gz','')
		CUR_DATABASE_DIR	= os.path.join(Data.DATABASE_DIR, DB_VERSION)
		PICKLE_VERSION		= open(os.path.join(CUR_DATABASE_DIR, 'VERSION')).readline().strip()
		OLD_DATABASE_PATH	= os.path.join(Data.DATABASE_DIR, 'old')
		IDS_DIR				= os.path.join(CUR_DATABASE_DIR, 'ids')
		REF_DIR				= os.path.join(CUR_DATABASE_DIR, 'databases')
		GTDB_DIR			= os.path.join(CUR_DATABASE_DIR, 'gtdb')
		PICKLE				= 'pickle'	
		HMM_SUFFIX 			= '.hmm'
		DMND_SUFFIX			= '.dmnd'
		KO_DB_NAME			= 'uniref100.KO'
		EC_DB_NAME			= 'uniref100.EC'
		PFAM_DB_NAME		= 'pfam'
		KO_HMM_DB_NAME		= 'ko'
		TIGRFAM_DB_NAME		= 'tigrfam'
		CAZY_DB_NAME		= 'cazy'
		GTDB_DB_NAME		= 'GTDB_R80_DB'
		
		GTDB_CAZY 			= os.path.join(GTDB_DIR, "gtdb_cazy.tsv")
		GTDB_KO 			= os.path.join(GTDB_DIR, "gtdb_ko.tsv")
		GTDB_PFAM 			= os.path.join(GTDB_DIR, "gtdb_pfam.tsv")
		GTDB_TIGRFAM 		= os.path.join(GTDB_DIR, "gtdb_tigrfam.tsv")
		GTDB_EC 			= os.path.join(GTDB_DIR, "gtdb_ec.tsv")

		TAXONOMY			= os.path.join(CUR_DATABASE_DIR, 'taxonomy_gtdb.tsv')
		M2DEF				= os.path.join(CUR_DATABASE_DIR, 'module_to_definition')
		M					= os.path.join(CUR_DATABASE_DIR, 'module_descriptions')
		COMPOUND_DESC		= os.path.join(CUR_DATABASE_DIR, 'br08001')    
		R2K					= os.path.join(CUR_DATABASE_DIR, 'reaction_to_orthology')
		R2C 				= os.path.join(CUR_DATABASE_DIR, 'reaction_to_compound')
		R2M					= os.path.join(CUR_DATABASE_DIR, 'reaction_to_module')
		M2R					= os.path.join(CUR_DATABASE_DIR, 'module_to_reaction')
		M2C					= os.path.join(CUR_DATABASE_DIR, 'module_to_cpd')
		R2P					= os.path.join(CUR_DATABASE_DIR, 'reaction_to_pathway')
		P2R					= os.path.join(CUR_DATABASE_DIR, 'pathway_to_reaction')
		C2R					= os.path.join(CUR_DATABASE_DIR, 'compound_to_reaction')
		C					= os.path.join(CUR_DATABASE_DIR, 'compound_descriptions')    
		R					= os.path.join(CUR_DATABASE_DIR, 'reaction_descriptions')
		P					= os.path.join(CUR_DATABASE_DIR, 'pathway_descriptions')
		K					= os.path.join(CUR_DATABASE_DIR, 'ko_descriptions')

		PFAM2CLAN			= os.path.join(CUR_DATABASE_DIR, 'pfam_to_clan')
		CLAN2NAME			= os.path.join(CUR_DATABASE_DIR, 'clan_to_name')
		PFAM2NAME			= os.path.join(CUR_DATABASE_DIR, 'pfam_to_name')
		PFAM2DESCRIPTION	= os.path.join(CUR_DATABASE_DIR, 'pfam_to_description')
		EC2DESCRIPTION	= os.path.join(CUR_DATABASE_DIR, 'ec_to_description')
		TIGRFAM2DESCRIPTION= os.path.join(CUR_DATABASE_DIR, 'tigrfam_descriptions')
		CLAN2PFAM			= os.path.join(CUR_DATABASE_DIR, 'clan_to_pfam')

	def __init__(self):

		self.signature_modules = set(['M00611', 'M00612', 'M00613', 'M00614',
		'M00617', 'M00618', 'M00615', 'M00616', 'M00363', 'M00542', 'M00574',
		'M00575', 'M00564', 'M00660', 'M00664', 'M00625', 'M00627', 'M00745',
		'M00651', 'M00652', 'M00704', 'M00725', 'M00726', 'M00730', 'M00744',
		'M00718', 'M00639', 'M00641', 'M00642', 'M00643', 'M00769', 'M00649',
		'M00696', 'M00697', 'M00698', 'M00700', 'M00702', 'M00714', 'M00705',
		'M00746'])

		logging.info("Loading databases")
		logging.debug("Loading module definitions")
		self.m2def = self.load_pickle(self.M2DEF)
		logging.debug("Loading module descriptions")
		self.m = self.load_pickle(self.M)
		logging.debug("Loading reaction to pathway information")
		self.r2p = self.load_pickle(self.R2P)
		logging.debug("Loading pathway to reaction information")
		self.p2r = self.load_pickle(self.P2R)
		logging.debug("Loading reaction to orthology information")
		self.r2k = self.load_pickle(self.R2K)
		logging.debug("Loading reaction to module information")
		self.r2m = self.load_pickle(self.R2M)
		logging.debug("Loading module to reaction information")
		self.m2r = self.load_pickle(self.M2R)
		logging.debug("Loading module to compound information")
		self.m2c = self.load_pickle(self.M2C)
		logging.debug("Loading reaction to compound information")
		self.r2c = self.load_pickle(self.R2C)
		logging.debug("Loading compound to reaction information")
		self.c2r = self.load_pickle(self.C2R)
		logging.debug("Loading compound descriptions")
		self.c = self.load_pickle(self.C)
		logging.debug("Loading pathway descriptions")
		self.p = self.load_pickle(self.P)
		logging.debug("Loading reaction descriptions")
		self.r = self.load_pickle(self.R)
		logging.debug("Loading ko descriptions")
		self.k = self.load_pickle(self.K)
		logging.debug("Loading compound classifications")
		self.compound_desc_dict = self.load_pickle(self.COMPOUND_DESC)
		logging.debug("Loading pfam to clan information")
		self.pfam2clan = self.load_pickle(self.PFAM2CLAN)
		logging.debug("Loading clan descriptions")
		self.clan2name = self.load_pickle(self.CLAN2NAME)
		logging.debug("Loading pfam names")
		self.pfam2name = self.load_pickle(self.PFAM2NAME)
		logging.debug("Loading pfam descriptions")
		self.pfam2description = self.load_pickle(self.PFAM2DESCRIPTION)
		logging.debug("Loading ec descriptions")
		self.ec2description = self.load_pickle(self.EC2DESCRIPTION)
		logging.debug("Loading pfam hierarchy")
		self.clan2pfam = self.load_pickle(self.CLAN2PFAM)
		logging.debug("Loading tigrfam descriptions")
		self.tigrfamdescription = self.load_pickle(self.TIGRFAM2DESCRIPTION)
		logging.info("Loading reference db paths")		


		self.k2r = dict()
		for reaction, kos in self.r2k.items():
			for ko in kos:
				if ko not in self.k2r:
					self.k2r[ko] = list()
				self.k2r[ko].append(reaction)

		self.taxonomy 		= self.parse_taxonomy(self.TAXONOMY)

		self.KO_DB 			= os.path.join(self.REF_DIR, self.KO_DB_NAME + self.DMND_SUFFIX)
		self.EC_DB 			= os.path.join(self.REF_DIR, self.EC_DB_NAME + self.DMND_SUFFIX)
		self.GTDB_DB		= os.path.join(self.REF_DIR, self.GTDB_DB_NAME)

		self.PFAM_DB 		= os.path.join(self.REF_DIR, self.PFAM_DB_NAME + self.HMM_SUFFIX)
		self.KO_HMM_DB 		= os.path.join(self.REF_DIR, self.KO_HMM_DB_NAME + self.HMM_SUFFIX)
		self.TIGRFAM_DB 	= os.path.join(self.REF_DIR, self.TIGRFAM_DB_NAME + self.HMM_SUFFIX)
		self.CAZY_DB 		= os.path.join(self.REF_DIR, self.CAZY_DB_NAME + self.HMM_SUFFIX)
		self.PFAM_CLAN_DB 	= os.path.join(self.IDS_DIR, 'PFAM_CLANS.txt')


	def load_pickle(self, file):

		with open('.'.join([file, self.PICKLE_VERSION, self.PICKLE]), 'rb') as file_io:
			loaded_pickle = pickle.load(file_io)

		return loaded_pickle

	def parse_taxonomy(self, taxonomy_path):
		
		output_taxonomy_dictionary = dict()

		for line in open(taxonomy_path):
			genome, taxonomy_string = line.strip().split('\t')
			output_taxonomy_dictionary[genome] = taxonomy_string.split(';')
			
		return output_taxonomy_dictionary



#!/usr/bin/env python3
###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

__author__      = "Joel Boyd"
__copyright__   = "Copyright 2017"
__credits__     = ["Joel Boyd"]
__license__     = "GPL3"
__version__     = "0.0.7"
__maintainer__  = "Joel Boyd"
__email__       = "joel.boyd near uq.net.au"
__status__      = "Development"
 
###############################################################################
# Imports
import logging
import os
from itertools import chain
from collections import Counter
# Local
from enrichm.databases import Databases
###############################################################################

class MatrixGenerator:
        
    KO      = 'KO_IDS.txt'
    EC      = 'EC_IDS.txt'
    PFAM    = 'PFAM_IDS.txt'
    TIGRFAM = 'TIGRFAM_IDS.txt'
    CAZY = 'CAZY_IDS.txt'
    HYPOTHETICAL = 'HYPOTHETICAL'

    def __init__(self, annotation_type, clusters = None):
        '''
        Interpret which annotation type to write a matrix for.

        Parameters
        ----------
        annotation_type - String.
        '''
        self.annotation_type = annotation_type

        if self.annotation_type == self.KO:
            self.annotation_list = [x.strip() for x in open(os.path.join(Databases.IDS_DIR, self.KO))]
        
        elif self.annotation_type == self.EC:
            self.annotation_list = [x.strip() for x in open(os.path.join(Databases.IDS_DIR, self.EC))]
        
        elif self.annotation_type == self.PFAM:
            self.annotation_list = [x.strip() for x in open(os.path.join(Databases.IDS_DIR, self.PFAM))]
        
        elif self.annotation_type == self.TIGRFAM:
            self.annotation_list = [x.strip() for x in open(os.path.join(Databases.IDS_DIR, self.TIGRFAM))]
        
        elif self.annotation_type == self.CAZY:
            self.annotation_list = [x.strip() for x in open(os.path.join(Databases.IDS_DIR, self.CAZY))]
        
        elif self.annotation_type == self.HYPOTHETICAL:
            self.annotation_list = clusters

    def write_matrix(self, genomes_list, count_domains, output_path):
        '''
        Writes a frequency matrix with of each annotation (rows) per sample (columns)

        Parameters
        ----------
        genomes_list        - list. List of Genome objects
        output_path         - string. Path to file to which the results are written.
        '''
        
        logging.info("    - Writing results to file: %s" % output_path)
        
        with open(output_path, 'w') as out_io:
            colnames = ['ID'] + [genome.name for genome in genomes_list]
            out_io.write('\t'.join(colnames) + '\n')
            
            if count_domains:
                genome_annotations = {genome.name:Counter(chain(*[sequence.all_annotations() for sequence in genome.sequences.values()]))
                                      for genome in genomes_list}
            else:
                genome_annotations = {genome.name:Counter(chain(*[set(sequence.all_annotations()) for sequence in genome.sequences.values()]))
                                      for genome in genomes_list}

            for annotation in self.annotation_list:
                output_line = [annotation]

                for genome in genomes_list:
                
                    if annotation in genome_annotations[genome.name]:
                        output_line.append(str(genome_annotations[genome.name][annotation]))
                
                    else:
                        output_line.append('0')
                
                out_io.write( '\t'.join(output_line) + '\n' )



#!/usr/bin/env python3
###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

__author__      = "Joel Boyd"
__copyright__   = "Copyright 2017"
__credits__     = ["Joel Boyd"]
__license__     = "GPL3"
__version__     = "0.0.7"
__maintainer__  = "Joel Boyd"
__email__       = "joel.boyd near uq.net.au"
__status__      = "Development"

###############################################################################
# Imports
import logging
import os
import pickle
import multiprocessing as mp
# Local
from enrichm.annotate import Annotate
################################################################################

def parse_genomes(path):
	logging.info('Loading: %s' % (os.path.basename(path)))
	genome = pickle.load(open(path, 'rb'))
	return genome

################################################################################

class ParseAnnotate:
	
	"""docstring for ParseAnnotate"""

	def __init__(self, enrichm_annotate_output, processes):
		self.path = enrichm_annotate_output
		# Parse genome objects
		self.genome_pickle_file_path \
			= os.path.join(enrichm_annotate_output, Annotate.GENOME_OBJ)
		self.processes \
			= processes

		ko = os.path.join(enrichm_annotate_output, Annotate.OUTPUT_KO)
		if os.path.isfile(ko):
			self.ko = ko
		else:
			self.ko = None
		pfam = os.path.join(enrichm_annotate_output, Annotate.OUTPUT_PFAM)
		if os.path.isfile(pfam):
			self.pfam = pfam
		else:
			self.pfam = None
		tigrfam = os.path.join(enrichm_annotate_output, Annotate.OUTPUT_TIGRFAM)
		if os.path.isfile(tigrfam):
			self.tigrfam = tigrfam
		else:
			self.tigrfam = None
		cazy = os.path.join(enrichm_annotate_output, Annotate.OUTPUT_CAZY)
		if os.path.isfile(cazy):
			self.cazy = cazy
		else:
			self.cazy = None
		ec = os.path.join(enrichm_annotate_output, Annotate.OUTPUT_EC)
		if os.path.isfile(ec):
			self.ec = ec
		else:
			self.ec = None
		hypothetical_cluster = os.path.join(enrichm_annotate_output, Annotate.OUTPUT_HYPOTHETICAL_CLUSTER)
		if os.path.isfile(hypothetical_cluster):
			self.hypothetical_cluster = hypothetical_cluster
		else:
			self.hypothetical_cluster = None
		hypothetical_ortholog = os.path.join(enrichm_annotate_output, Annotate.OUTPUT_HYPOTHETICAL_ORTHOLOG)
		if os.path.isfile(hypothetical_ortholog):
			self.hypothetical_ortholog = hypothetical_ortholog
		else:
			self.hypothetical_ortholog = None

	def parse_pickles(self, path, genome_list):
		'''
		Opens the pickled genome objects from a previous run of enrichm 
		annotate

		Parameters
		----------
		enrichm_annotate_output - String. Output directory of a previous run 
								  of enrichm annotate (At lease version 0.0.7)
								  or above
		Outputs
		-------
		List of Genome objects
		'''	
		logging.info('Loading genome pickles')
		
		output_genome_list 	= list()
		paths 				= list()
		


		for pickled_genome in genome_list:
			pickled_genome_path = os.path.join(path, pickled_genome + '.pickle')
			if os.path.isfile(pickled_genome_path):
				paths.append(pickled_genome_path)
		
		self.pool = mp.Pool(processes = self.processes)
		output_genome_list = self.pool.map_async(parse_genomes, paths)
		output_genome_list.wait()
		genome_objects = output_genome_list.get()
		self.pool.close()
		return genome_objects